// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

#include "gtest/gtest.h"
#include "test/providers/provider_test_utils.h"

namespace onnxruntime {
namespace test {

template <typename T>
struct MatMulTestData {
  std::string name;
  std::vector<int64_t> input0_dims;
  std::vector<int64_t> input1_dims;
  std::vector<int64_t> expected_dims;
  std::vector<T> expected_vals;
};

template <typename T>
std::vector<MatMulTestData<T>> GenerateTestCases()
{
  std::vector<MatMulTestData<T>> test_cases;

  test_cases.push_back(
    {"test padding and broadcast",
    {3, 1, 1, 2},
    {2, 2, 2},
    {3, 2, 1, 2},
    {2, 3, 6, 7, 6, 11, 26, 31, 10, 19, 46, 55}});

  test_cases.push_back(
    {"test padding and broadcast",
    {2, 3, 2},
    {3, 2, 2, 1},
    {3, 2, 3, 1},
    {1, 3, 5, 33, 43, 53, 5, 23, 41, 85, 111, 137, 9, 43, 77, 137, 179, 221}});

  test_cases.push_back(
    {"test left 1D",
    {2},
    {3, 2, 1},
    {3, 1},
    {1, 3, 5}});

  test_cases.push_back(
    {"test right 1D",
    {3, 1, 2},
    {2},
    {3, 1},
    {1, 3, 5}});

  test_cases.push_back(
    {"test scalar output",
    {3},
    {3},
    {},
    {5}});

  test_cases.push_back(
    {"test 2D",
    {3, 4},
    {4, 3},
    {3, 3},
    {42, 48, 54, 114, 136, 158, 186, 224, 262}});

  test_cases.push_back(
    {"test 2D special",
    {2, 2, 3},
    {3, 4},
    {2, 2, 4},
    {20, 23, 26, 29, 56, 68, 80, 92, 92, 113, 134, 155, 128, 158, 188, 218}});

  test_cases.push_back(
    {"test 2D special 2",
    {2, 2, 3},
    {1, 3, 4},
    {2, 2, 4},
    {20, 23, 26, 29, 56, 68, 80, 92, 92, 113, 134, 155, 128, 158, 188, 218}});

  return test_cases;
}

template <typename T>
void RunMatMulTest(int32_t opset_version = 7)
{
  std::vector<T> common_input_vals{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11};
  for (auto t : GenerateTestCases<T>()) {
    OpTester test("MatMul", opset_version);

    int64_t size0 = TensorShape::ReinterpretBaseType(t.input0_dims).SizeHelper(0, t.input0_dims.size());
    std::vector<T> input0_vals(common_input_vals.cbegin(), common_input_vals.cbegin() + size0);
    test.AddInput<T>("A", t.input0_dims, input0_vals);

    int64_t size1 = TensorShape::ReinterpretBaseType(t.input1_dims).SizeHelper(0, t.input1_dims.size());
    std::vector<T> input1_vals(common_input_vals.cbegin(), common_input_vals.cbegin() + size1);
    test.AddInput<T>("B", t.input1_dims, input1_vals);

    test.AddOutput<T>("Y", t.expected_dims, t.expected_vals);

    test.Run(OpTester::ExpectResult::kExpectSuccess, "", {kTensorrtExecutionProvider});// Disable TensorRT because of unsupported data type
  }
}

TEST(MathOpTest, MatMulFloatType) {
  RunMatMulTest<float>(7);
}

TEST(MathOpTest, MatMulDoubleType) {
  RunMatMulTest<double>(7);
}

TEST(MathOpTest, MatMulInt32Type) {
  RunMatMulTest<int32_t>(9);
}

TEST(MathOpTest, MatMulUint32Type) {
  RunMatMulTest<uint32_t>(9);
}

TEST(MathOpTest, MatMulInt64Type) {
  RunMatMulTest<int64_t>(9);
}

TEST(MathOpTest, MatMulUint64Type) {
  RunMatMulTest<uint64_t>(9);
}

}  // namespace test
}  // namespace onnxruntime
